//====================================================================================================================
// Get OS name
//====================================================================================================================
static char *GetOsName ()
{
	static char		szWinVista[] = "Vista";
	static char		szWin2k3[] = "Server 2003";
	static char		szWinXP[] = "XP";
	static char		szWin2k[] = "2000";
	static char		szWinMe[] = "Me";
	static char		szWin98[] = "98";
	static char		szWin95[] = "95";
	static char		szWinNT4[] = "NT 4.0";
	static char		szWinNT3[] = "NT 3.51";
	static char		szWinNT[] = "NT";
	static char		szWindows[128];
	char					*szName;

	OSVERSIONINFO ovi;
	ovi.dwOSVersionInfoSize = sizeof (ovi);
	if (GetVersionEx (&ovi))
	{
		szName = (ovi.dwPlatformId == VER_PLATFORM_WIN32_NT) ? szWinNT : NULL;
		switch (ovi.dwMajorVersion)
		{
		case 6:
			switch (ovi.dwMinorVersion)
			{
			case 0:
				szName = szWinVista;
				break;
			}
			break;
		case 5:
			switch (ovi.dwMinorVersion)
			{
			case 2:
				szName = szWin2k3;
				break;
			case 1:
				szName = szWinXP;
				break;
			case 0:
				szName = szWin2k;
				break;
			}
			break;
		case 4:
			switch (ovi.dwMinorVersion)
			{
			case 90:
				szName = szWinMe;
				break;
			case 10:
				szName = szWin98;
				break;
			case 0:
				szName = (ovi.dwPlatformId == VER_PLATFORM_WIN32_NT) ? szWinNT4 : szWin95;
				break;
			}
			break;
		case 3:
			if (ovi.dwMinorVersion == 51) szName = szWinNT3;
			break;
		}

		wsprintf (szWindows, "Windows %s build %i %s", szName, ovi.dwBuildNumber, ovi.szCSDVersion);
		return (szWindows);
	}

	return (NULL);
}

//====================================================================================================================
// Get OS type
//====================================================================================================================
static char *GetOsType ()
{
	static char		szWin32s[] = "Win32s";
	static char		szWin9x[] = "Windows 9x";
	static char		szWinNT[] = "Windows NT";
	OSVERSIONINFO ovi;
	ovi.dwOSVersionInfoSize = sizeof (ovi);
	if (GetVersionEx (&ovi))
	{
		switch (ovi.dwPlatformId)
		{
		case VER_PLATFORM_WIN32_NT:
			return (szWinNT);
		case VER_PLATFORM_WIN32_WINDOWS:
			return (szWin9x);
		case VER_PLATFORM_WIN32s:
			return (szWin32s);
		}
	}

	return (NULL);
}

static int		curlength = 0;
static int		fullength = 0;
static char		*output = NULL;

static bool		RunFlag = false;
static void		*StdOutput = NULL;
static HANDLE PipeStdOutputDup = NULL;
static HANDLE StdOutputThread = NULL;
static DWORD WINAPI StdOutputThreadFunction (LPVOID lpParameter)
{
	BOOL	OkFlag;
	CHAR	Buffer[4096];
	DWORD ReadCount;
	while (!RunFlag) Sleep (1);
	while (true)
	{
		OkFlag = ReadFile (PipeStdOutputDup, Buffer, 4096, &ReadCount, NULL);
		if ((OkFlag) && (ReadCount > 0))
		{
			if (curlength + ReadCount > fullength - 1)
			{
				char	*temp = new char[fullength + MAX_STR_LEN];
				memcpy (temp, output, curlength);
				delete[] output;
				output = temp;
				fullength += MAX_STR_LEN;
			}

			for (DWORD i = 0; i < ReadCount; i++) output[curlength + i] = Buffer[i];
			curlength += ReadCount;
		}
		else
			break;
		Sleep (1);
	}

	return (0);
}

static void ParseExec (char *pf)
{
	int i;
	if (!pf) return ;
	output = new char[MAX_STR_LEN];
	curlength = 0;
	fullength = MAX_STR_LEN;
	StdOutput = GetStdHandle (STD_OUTPUT_HANDLE);

	BOOL							CreateStdOutput;
	HANDLE							PipeStdOutputWrite, PipeStdOutputRead;
	SECURITY_ATTRIBUTES saAttrStdOutput;
	saAttrStdOutput.nLength = sizeof (SECURITY_ATTRIBUTES);
	saAttrStdOutput.bInheritHandle = true;
	saAttrStdOutput.lpSecurityDescriptor = NULL;
	CreateStdOutput = CreatePipe (&PipeStdOutputRead, &PipeStdOutputWrite, &saAttrStdOutput, 0);
	CreateStdOutput = SetStdHandle (STD_OUTPUT_HANDLE, PipeStdOutputWrite);
	CreateStdOutput = DuplicateHandle
		(
			GetCurrentProcess (),
			PipeStdOutputRead,
			GetCurrentProcess (),
			&PipeStdOutputDup,
			0,
			FALSE,
			DUPLICATE_SAME_ACCESS
		);
	CreateStdOutput = CloseHandle (PipeStdOutputRead);

	DWORD StdOutputThreadID;
	StdOutputThread = CreateThread
		(
			NULL,
			0,
			(LPTHREAD_START_ROUTINE) StdOutputThreadFunction,
			NULL,
			0,
			&StdOutputThreadID
		);

	STARTUPINFOA				si;
	PROCESS_INFORMATION pi;
	ZeroMemory (&si, sizeof (si));
	ZeroMemory (&pi, sizeof (pi));
	si.cb = sizeof (si);
	si.dwFlags = STARTF_USESTDHANDLES;
	si.hStdOutput = GetStdHandle (STD_OUTPUT_HANDLE);
	si.hStdInput = GetStdHandle (STD_INPUT_HANDLE);
	si.hStdError = GetStdHandle (STD_ERROR_HANDLE);
	if (CreateProcess (NULL, pf, NULL, NULL, TRUE, NORMAL_PRIORITY_CLASS, NULL, NULL, &si, &pi) == TRUE)
	{
		RunFlag = true;
		WaitForSingleObject (pi.hProcess, INFINITE);
		CloseHandle (pi.hProcess);
		CloseHandle (pi.hThread);
	}

	HANDLE	CloseStdOutput;
	CloseStdOutput = GetStdHandle (STD_OUTPUT_HANDLE);
	CloseHandle (CloseStdOutput);
	SetStdHandle (STD_OUTPUT_HANDLE, StdOutput);
	WaitForSingleObject (StdOutputThread, INFINITE);
	CloseHandle (PipeStdOutputDup);

	//CharToOemBuff(output,output,curlength);
	for (i = 0; i < curlength; i++)
	{
		if (output[i] == 0x0A)
		{
			for (int j = i; j < curlength - 1; j++) output[j] = output[j + 1];
			curlength--;
			output[curlength] = '\0';
		}
	}

	EditorInsertText (output);
	for (i = 0; i < curlength; i++)
		if ((i > 0) && (output[i - 1] == '\0')) EditorInsertText (output + i);
	delete[] output;
}

static char *GetParamParam (char *szParam)
{
	if (szParam)
	{
		char	*p = szParam;
		char	*o = szParam;
		while (*p)
		{
			if ((*p == '\\') && ((p[1] == '\\') || (p[1] == ':') || (p[1] == '@')))
				p++;
			else if (*p == ':')
			{
				*o = 0;
				return (p + 1);
			}

			if (*p) *o++ = *p++;
		}

		*o = 0;
	}

	return (NULL);
}

static char *DoCounter (char *szName, char *szParam, char *szValue)
{
	static char sz0[] = "0";
	if (szParam)
	{
		char	*szParamInt = (*szParam) ? szParam : sz0;
		WritePrivateProfileString ("COUNTERS", szName, szParamInt, szIni);
		return (szParamInt);
	}
	else
	{
		FSF.itoa (GetPrivateProfileInt ("COUNTERS", szName, 0, szIni), szValue, 10);
		return (szValue);
	}
}

static char *DoString (char *szName, char *szParam, char *szValue, DWORD dLength)
{
	if (szParam)
	{
		WritePrivateProfileString ("STRINGS", szName, szParam, szIni);
		return (szParam);
	}
	else
	{
		GetPrivateProfileString ("STRINGS", szName, "", szValue, dLength, szIni);
		return (szValue);
	}
}

static void ParseFile (char *pf, int *setPos, TEditorPos *pos);

static void DoCommand (TCOMMAND eCmd, char *szParam)
{
	const char		fmtTime[] = "HH:mm:ss";
	const char		fmtDate[] = "dd.MM.yyyy";
	const char		fmtDateTime[] = "HH:mm:ss dd.MM.yyyy";
	char					szDate[128];
	char					szString[10240] = "";
	char					*psz = NULL, *ppsz, *szParam1, *szParam2, *szParam3;
	bool					conv = true;
	bool					toup = false;
	unsigned long l;
	szParam1 = szParam;
	szParam2 = GetParamParam (szParam1);
	switch (eCmd)
	{
	case CMD_EditorString:
		{
			EditorGetString egs;
			EditorGetStr (&egs);
			psz = lstrcpyn (szString, egs.StringText, egs.StringLength + 1);
		}
		break;
	case CMD_EditorPos:
		{
			TEditorPos	pos = EditorGetPos ();
			FSF.itoa (pos.Row + 1, szString, 10);
			psz = szString;
		}
		break;
	case CMD_EditorCol:
		{
			TEditorPos	pos = EditorGetPos ();
			FSF.itoa (pos.Col + 1, szString, 10);
			psz = szString;
		}
		break;
	case CMD_Argument:
		{
			psz = userString[*szParam1 - 0x30];
		}
		break;
	case CMD_ClipBoard:
		{
			ppsz = FSF.PasteFromClipboard ();
			if (ppsz)
			{
				psz = (char *) malloc (lstrlen (ppsz) + 1);
				lstrcpy (psz, ppsz);
				FSF.DeleteBuffer (ppsz);
			}

			conv = false;
		}
		break;
	case CMD_Selected:
		{
			EditorSaveRestore ();
		}
		break;
	case CMD_CompName:
		{
			psz = szString;
			l = sizeof (szString);
			GetComputerName (szString, &l);
		}
		break;
	case CMD_Counter:
		{
			if (szParam1 && *szParam1)
			{
				szParam3 = GetParamParam (szParam2);
				psz = DoCounter (szParam1, szParam2, szString);
			}
		}
		break;
	case CMD_Date:
		{
			psz = (szParam1) ? szParam1 : (char *) fmtDate;
		}
		break;
	case CMD_DateTime:
		{
			psz = (szParam1) ? szParam1 : (char *) fmtDateTime;
		}
		break;
	case CMD_Decrement:
		{
			szParam3 = GetParamParam (szParam2);
			DoCounter (szParam1, NULL, szString);
			FSF.itoa (FSF.atoi (szString) - ((szParam2) ? FSF.atoi (szParam2) : 1), szString, 10);
			psz = DoCounter (szParam1, szString, NULL);
		}
		break;
	case CMD_Exec:
		{
			ParseExec (szParam1);
		}
		break;
	case CMD_Enviroment:
		{
			psz = szString;
			GetEnvironmentVariable (szParam1, szString, sizeof (szString));
		}
		break;
	case CMD_File:
	case CMD_FileUp:
		{
			psz = lstrcpy (szString, GetFilePathName ());
			toup = (eCmd == CMD_FileUp);
			conv = false;
		}
		break;
	case CMD_FileExt:
	case CMD_FileExtUp:
		{
			psz = lstrcpy (szString, Point2FileExt (GetFilePathName ()));
			toup = (eCmd == CMD_FileExtUp);
			conv = false;
		}
		break;
	case CMD_FileName:
	case CMD_FileNameUp:
		{
			psz = lstrcpy (szString, Point2FileName (GetFilePathName ()));
			*(Point2FileExt (psz) - 1) = 0;
			toup = (eCmd == CMD_FileNameUp);
			conv = false;
		}
		break;
	case CMD_FilePath:
	case CMD_FilePathUp:
		{
			psz = lstrcpy (szString, GetFilePathName ());
			*Point2FileName (psz) = 0;
			toup = (eCmd == CMD_FilePathUp);
			conv = false;
		}
		break;
	case CMD_FileNameExt:
	case CMD_FileNameExtUp:
		{
			psz = lstrcpy (szString, Point2FileName (GetFilePathName ()));
			toup = (eCmd == CMD_FileNameExtUp);
			conv = false;
		}
		break;
	case CMD_GUID:
		{
			GUID	Guid;
			char	*pw = psz = szString;
			CoInitialize (NULL);
			CoCreateGuid (&Guid);
			StringFromGUID2 (Guid, (LPOLESTR) szString, sizeof (szString));
			while (true)
			{
				*psz++ = *pw++;
				if (!* ++pw) break;
			}

			*psz = 0;
			psz = szString;
			CoUninitialize ();
		}
		break;
	case CMD_Increment:
		{
			szParam3 = GetParamParam (szParam2);
			DoCounter (szParam1, NULL, szString);
			FSF.itoa (FSF.atoi (szString) + ((szParam2) ? FSF.atoi (szParam2) : 1), szString, 10);
			psz = DoCounter (szParam1, szString, NULL);
		}
		break;
	case CMD_Input:
	case CMD_InputString:
		{
			char	*szParam4;
			if (szParam1 && *szParam1)
				szParam3 = GetParamParam (szParam2);
			else
				break;
			if (eCmd == CMD_InputString)
			{
				if (szParam2 && *szParam2)
					szParam4 = GetParamParam (szParam3);
				else
					break;
				DoString (szParam1, NULL, szString, sizeof (szString));
			}

			char	*p1 = (eCmd != CMD_InputString) ? szParam1 : szParam2;
			char	*p2 = (eCmd != CMD_InputString) ? szParam2 : ((szParam3) ? szParam3 : szString);
			struct InitDialogItemEx InitItems[] =
			{															//Type X1 Y1 X2 Y2 Fo Se Fl DB Data
																		///;
																		///00
				{ DI_DOUBLEBOX, 3, 1, 55, 6, 0, 0, DIF_BOXCOLOR, 0, GetMsg (MInput), NULL },

				//01
				{ DI_TEXT, 5, 2, 0, 0, 0, 0, 0, 0, p1, NULL },

				//02
				{ DI_EDIT, 5, 3, 53, 0, 1, 0, DIF_HISTORY, 0, p2, "True-Tpl.History.UserInput" },

				//03
				{ DI_TEXT, 0, 4, 0, 0, 0, 0, DIF_BOXCOLOR | DIF_SEPARATOR, 0, "", NULL },

				//04
				{ DI_BUTTON, 0, 5, 0, 0, 0, 0, DIF_CENTERGROUP, 1, GetMsg (MOK), NULL },

				//05
				{ DI_BUTTON, 0, 5, 0, 0, 0, 0, DIF_CENTERGROUP, 0, GetMsg (MCancel), NULL }
			};
			struct FarDialogItem		DialogItems[(sizeof (InitItems) / sizeof (InitItems[0]))];
			InitDialogItemsEx (InitItems, DialogItems, (sizeof (InitItems) / sizeof (InitItems[0])));

			int n = Info.Dialog
				(
					Info.ModuleNumber,
					-1,
					-1,
					59,
					8,
					NULL,
					DialogItems,
					(sizeof (InitItems) / sizeof (InitItems[0]))
				);
			if (n == 4)
			{
				if (eCmd != CMD_InputString)
				{
					EditorInsertText (DialogItems[2].Data);

					char	*szParam4;
					if (szParam2 && *szParam2)
					{
						szParam4 = GetParamParam (szParam3);
						if (szParam3) DoString (szParam3, DialogItems[2].Data, NULL, 0);
					}
				}
				else
					DoString (szParam1, DialogItems[2].Data, NULL, 0);
			}
		}
		break;
	case CMD_OsName:
		{
			psz = GetOsName ();
		}
		break;
	case CMD_OsType:
		{
			psz = GetOsType ();
		}
		break;
	case CMD_Random:
		{
			psz = szString;
			wsprintf (szString, "%i", rand ());
		}
		break;
	case CMD_String:
		{
			if (szParam1 && *szParam1)
			{
				szParam3 = GetParamParam (szParam2);
				psz = DoString (szParam1, szParam2, szString, sizeof (szString));
			}

			conv = false;
		}
		break;
	case CMD_Ticks:
		{
			psz = szString;
			wsprintf (szString, "%i", GetTickCount ());
		}
		break;
	case CMD_Time:
		{
			psz = (szParam1) ? szParam1 : (char *) fmtTime;
		}
		break;
	case CMD_UserName:
		{
			psz = szString;
			l = sizeof (szString);
			GetUserName (szString, &l);
		}
		break;
	}

	if ((eCmd == CMD_Date) || (eCmd == CMD_Time) || (eCmd == CMD_DateTime))
	{
		SYSTEMTIME	SysDate;
		GetLocalTime (&SysDate);
		if (eCmd != CMD_Date)
		{
			GetTimeFormat (LOCALE_SYSTEM_DEFAULT, 0, &SysDate, psz, szString, sizeof (szString));
			psz = szString;
		}

		if (eCmd != CMD_Time)
		{
			GetDateFormat (LOCALE_SYSTEM_DEFAULT, 0, &SysDate, psz, szDate, sizeof (szDate));
			psz = szDate;
		}
	}

	if (psz)
	{
		char	*str = psz;
		if (conv) CharToOem (psz, psz);
		if (toup) FSF.LUpperBuf (str, lstrlen (str));
		while (*psz)
		{
			if (*psz == 0x0A)
			{
				*psz = 0;										//if (str!=psz)
				EditorInsertText (str);
				str = ++psz;
			}
			else
				psz++;
		}

		if (*str) EditorInsertText (str);
	}

	if (eCmd == CMD_ClipBoard)
	{
		if (psz) free (psz);
	}																	//free allocated for this case memory
}

static void ParseFile (char *pf, int *setPos, TEditorPos *pos)
{
	static int	iCnt[10];
	char				szFilePath[_MAX_PATH + 1];
	char				szFileName[_MAX_FNAME + 1];
	TCOMMAND		eCommand;
	char				*f = pf, *fn = szFilePath;
	if (f && *f)
	{
		if ((*f++ == '$') && (*f == '\\'))
		{
			f++;
			GetModuleFileName (hInst, szFilePath, _MAX_PATH);
			fn = Point2FileName (szFilePath);
		}
		else
			f--;
		lstrcpy (fn, f);
		lstrcpy (szFileName, Point2FileName (szFilePath));
		*Point2FileName (szFilePath) = 0;

		char	*cBuffer = getFile (szFilePath, szFileName);
		if (cBuffer)
		{
			char	*s = f = cBuffer;
			char	*macro, *param, *next;
			while (*f)
			{
				switch (*f)
				{
				case 0x0A:
					{
						*f = 0;
						if (s != f) EditorInsertText (s);
						s = ++f;
						break;
					}

				case cMD:
					{
						if (*++f == cMD)
						{
							f++;
							if (*f == cMD)
							{
								f[-1] = '\0';
								if (s != f) EditorInsertText (s);
								f[-1] = cMD;
								s = --f;
								break;
							}

							bool	blParsed;
							bool	blSetPos = false;
							byte	isDateTime = 0;
							if (lstrlen (f) > 5)	//min_para_size==4 + 2 delimiters
							{
								char	_d;
								next = f;
								macro = NULL;
								param = NULL;
								while (*next)
								{
									if (*next == '\\')
										next++;
									else
									{
										if ((*next == ':') && !param) param = next + 1;
										if ((*next == cMD) && (next[1] == cMD))
										{
											macro = f;
											*next = 0;
											break;
										}
									}

									if (*next) next++;
								}

								if (macro)
								{
									eCommand = CMD_DoNothing;
									if (param)
									{
										_d = param[-1];
										param[-1] = 0;
									}

									if (!lstrcmpi2 (macro, "ARG"))
									{
										unsigned	c = macro[3];
										if ((c < '0') || (c > '9'))
											eCommand = CMD_DoNothing;
										else
										{
											param = macro + 3;
											eCommand = CMD_Argument;
										}
									}
									else if (!lstrcmpi (macro, "EDITOR_STRING"))
										eCommand = CMD_EditorString;
									else if (!lstrcmpi (macro, "EDITOR_POS"))
										eCommand = CMD_EditorPos;
									else if (!lstrcmpi (macro, "EDITOR_COL"))
										eCommand = CMD_EditorCol;
									else if (!lstrcmpi (macro, "CNT"))
										eCommand = CMD_Counter;
									else if (!lstrcmpi (macro, "STR"))
										eCommand = CMD_String;
									else if (!lstrcmpi (macro, "INC"))
										eCommand = CMD_Increment;
									else if (!lstrcmpi (macro, "DEC"))
										eCommand = CMD_Decrement;
									else if (!lstrcmpi (macro, "HERE"))
									{
										blSetPos = true;
									}
									else if (!lstrcmpi (macro, "DATE"))
										eCommand = CMD_Date;
									else if (!lstrcmpi (macro, "TIME"))
										eCommand = CMD_Time;
									else if (!lstrcmpi (macro, "DATETIME"))
										eCommand = CMD_DateTime;
									else if (!lstrcmpi (macro, "TICKS"))
										eCommand = CMD_Ticks;
									else if (!lstrcmpi (macro, "RAND"))
										eCommand = CMD_Random;
									else if (!lstrcmpi (macro, "INPUT"))
										eCommand = CMD_Input;
									else if (!lstrcmpi (macro, "INPUTSTR"))
										eCommand = CMD_InputString;
									else if (!lstrcmpi (macro, "INCLUDE"))
									{
										GetParamParam (param);
										ParseFile (param, setPos, pos);
									}
									else if (!lstrcmpi (macro, "EXEC"))
										eCommand = CMD_Exec;
									else if (!lstrcmpi (macro, "CLIP"))
										eCommand = CMD_ClipBoard;
									else if (!lstrcmpi (macro, "SELECTED"))
										eCommand = CMD_Selected;
									else if (!lstrcmpi (macro, "ENV"))
										eCommand = CMD_Enviroment;
									else if (!lstrcmpi (macro, "COMP_NAME"))
										eCommand = CMD_CompName;
									else if (!lstrcmpi (macro, "USER_NAME"))
										eCommand = CMD_UserName;
									else if (!lstrcmpi (macro, "OS_NAME"))
										eCommand = CMD_OsName;
									else if (!lstrcmpi (macro, "OS_TYPE"))
										eCommand = CMD_OsType;
									else if (!lstrcmpi (macro, "GUID"))
										eCommand = CMD_GUID;
									else if (!lstrcmpi2 (macro, "FILE_"))
									{
										macro += 5;
										if (!strcmpi (macro, "PATH_NAME"))
											eCommand = CMD_File;
										else if (!strcmpi (macro, "PATH"))
											eCommand = CMD_FilePath;
										else if (!strcmpi (macro, "NAME"))
											eCommand = CMD_FileName;
										else if (!strcmpi (macro, "EXT"))
											eCommand = CMD_FileExt;
										else if (!strcmpi (macro, "NAME_EXT"))
											eCommand = CMD_FileNameExt;
										else
											eCommand = CMD_Unknown;
										macro -= 5;
									}
									else if (!lstrcmpi2 (macro, "UPPERFILE_"))
									{
										macro += 10;
										if (!strcmpi (macro, "PATH_NAME"))
											eCommand = CMD_FileUp;
										else if (!strcmpi (macro, "PATH"))
											eCommand = CMD_FilePathUp;
										else if (!strcmpi (macro, "NAME"))
											eCommand = CMD_FileNameUp;
										else if (!strcmpi (macro, "EXT"))
											eCommand = CMD_FileExtUp;
										else if (!strcmpi (macro, "NAME_EXT"))
											eCommand = CMD_FileNameExtUp;
										else
											eCommand = CMD_Unknown;
										macro -= 10;
									}
									else
										eCommand = CMD_Unknown;
									if (eCommand == CMD_Unknown)
									{
										if (param) param[-1] = _d;
										*next = cMD;
										break;
									}
									else
									{
										macro[-2] = 0;
										if (*s) EditorInsertText (s);
										f = next + 2;
										s = f;
										if (blSetPos)
										{
											*setPos = 1;
											*pos = EditorGetPos ();
											pos->LeftCol = 0;
										}

										if (eCommand != CMD_DoNothing) DoCommand (eCommand, param);
									}
								}
							}
						}
						break;
					}

				default:
					f++;
				}
			}

			EditorInsertText (s);
			delete[] cBuffer;
		}
	}
}

static void RunMacro (TMacro *m, const char *origStr, int bounds[][2])
{
	bool	firstsel = false;
	for (char *p = m->MacroText; *p; p++)
	{
		if (*p == '\\')
		{
			switch (*++p)
			{
			case 'l':
				firstsel = true;
				break;
			}
		}
	}

	if (firstsel)
	{
		EditorInfo	ei;
		Info.EditorControl (ECTL_GETINFO, &ei);
		if (ei.BlockType != BTYPE_NONE) EditorSaveSelected ();
		Info.EditorControl (ECTL_DELETEBLOCK, NULL);
	}

	if (scanUserInput (true, '\\', m->MacroText))
	{
		int					userStrN, insPos = 0, setPos = 0, setStart = 0, setEnd = 0;
		bool				makeTime = false;
		char				ins[MAX_STR_LEN] = "";
		TEditorPos	pos, curPos, selStart, selEnd;
		pluginBusy = 1;
		for (char *p = m->MacroText; *p; p++)
		{
			char	*cb;
			if (*p == '\\')
			{
				insIns (ins, insPos);
				switch (*++p)
				{
				case 'p':
					setPos = 1;
					pos = EditorGetPos ();
					break;
				case 's':
					setStart = 1;
					selStart = EditorGetPos ();
					break;
				case 'e':
					setEnd = 1;
					selEnd = EditorGetPos ();
					break;
				case 0:
				case '\\':
					EditorProcessFARKey (KEY_BACKSLASH);
					break;
				case 't':
					EditorProcessFARKey (KEY_TAB);
					break;
				case 'b':
					EditorProcessFARKey (KEY_BS);
					break;
				case '^':
					EditorProcessFARKey (KEY_UP);
					break;
				case 'v':
					EditorProcessFARKey (KEY_DOWN);
					break;
				case '<':
					EditorProcessFARKey (KEY_LEFT);
					break;
				case '>':
					EditorProcessFARKey (KEY_RIGHT);
					break;
				case '[':
					EditorProcessFARKey (KEY_HOME);
					break;
				case ']':
					EditorProcessFARKey (KEY_END);
					break;
				case '(':
					EditorProcessFARKey (KEY_LEFT | KEY_CTRL);
					break;
				case ')':
					EditorProcessFARKey (KEY_RIGHT | KEY_CTRL);
					break;
				case 'r':
					EditorProcessReturnKey (0, 1);
					break;
				case 'n':
					EditorProcessReturnKey ();
					break;
				case '!':
					pluginBusy = 0;
					EditorProcessKey (*++p);
					pluginBusy = 1;
				case '?':
					p++;
					while (*p)
					{
						bool	inname = false;
						if ((p[-1] != '\\') && (p[0] == '\'')) inname = !inname;
						if ((!inname) && (p[-1] != '\\') && (p[0] == '?'))
							break;
						else
							p++;
					}
					break;
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					userStrN = *p - '0';
					EditorInsertText (userString[userStrN]);
					break;
				case '$':
					switch (p[1])
					{
					case '0':
					case '1':
					case '2':
					case '3':
					case '4':
					case '5':
					case '6':
					case '7':
					case '8':
					case '9':
						{
							p++;

							char	*ss = makeSubstr (*p - '0' + 1, origStr, bounds, 11);
							if (ss)
							{
								EditorInsertText (ss);
								free (ss);
							}
						}
						break;
					default:
						EditorProcessKey (*p);
						break;
					}
					break;
				case '{':
					if ((cb = lstrchr (p + 1, '}')) == NULL)
						EditorProcessKey (*p);
					else
					{
						char	ch = *cb;
						*cb = 0;
						EditorProcessGlobalFARKey (FSF.FarNameToKey (p + 1));
						*cb = ch;
						p = cb;
					}
					break;
				case 'w':
					makeTime = true;
				case 'd':
					{
						char	format[80] = "", *fmt = NULL;
						if (p[1] == '\'')
						{
							char	*t = format;
							p++;
							while (*++p && *p != '\'') *t++ = (*p == '`') ? '\'' : *p;
							*t = 0;
							fmt = format;
						}

						if (makeTime)
							DoCommand (CMD_Time, fmt);
						else
							DoCommand (CMD_Date, fmt);
						break;
					}

				case 'c':
					{
						DoCommand (CMD_ClipBoard, NULL);
						break;
					}

				case 'l':
					{
						DoCommand (CMD_Selected, NULL);
						break;
					}

				case 'f':
					{
						p++;
						if (*p == 'f')
							DoCommand (CMD_File, NULL);
						else if (*p == 'p')
							DoCommand (CMD_FilePath, NULL);
						else if (*p == 'n')
							DoCommand (CMD_FileName, NULL);
						else if (*p == 'e')
							DoCommand (CMD_FileExt, NULL);
						else if (*p == 'x')
							DoCommand (CMD_FileNameExt, NULL);
						break;
					}

				case 'F':
					{
						p++;
						if (*p == 'F')
							DoCommand (CMD_FileUp, NULL);
						else if (*p == 'P')
							DoCommand (CMD_FilePathUp, NULL);
						else if (*p == 'N')
							DoCommand (CMD_FileNameUp, NULL);
						else if (*p == 'E')
							DoCommand (CMD_FileExtUp, NULL);
						else if (*p == 'X')
							DoCommand (CMD_FileNameExtUp, NULL);
						break;
					}

				case 'E':
					{
						p++;
						if (*p == 's')
							DoCommand (CMD_EditorString, NULL);
						else if (*p == 'p')
							DoCommand (CMD_EditorPos, NULL);
						else if (*p == 'c')
							DoCommand (CMD_EditorCol, NULL);
						break;
					}

				case 'u':
					{
						DoCommand (CMD_UserName, NULL);
						break;
					}

				case 'i':
					{
						ParseFile (GetXmlParam (&p), &setPos, &pos);
						break;
					}

				case 'o':
					{
						ParseExec (GetXmlParam (&p));
						break;
					}

				default:
					pluginBusy = 0;
					EditorProcessKey (*p);
					pluginBusy = 1;
					break;
				}
			}
			else
			{
				ins[insPos++] = *p;
				ins[insPos] = 0;
			}
		}

		insIns (ins, insPos);
		pluginBusy = 0;
		if (setPos) EditorSetPos (pos);
		if (setStart && setEnd) selectBlock (selStart.Col, selStart.Row, selEnd.Col, selEnd.Row);
	}

	if (firstsel) EditorSaveRemove ();
}
