struct TCompiler
{
	TCompiler();
	TCompiler (const TCompiler &);
	char	title[MAX_REG_LEN], err[MAX_STR_LEN];
	int		line, col, fileMatch;
	bool	searchCol;
	void defaults (void)
	{
		*title = *err = 0;
		line = col = fileMatch = -1;
		searchCol = false;
	}
};

TCompiler::TCompiler ()
{
	defaults ();
}

TCompiler::TCompiler (const TCompiler &e)
{
	lstrcpy (title, e.title);
	lstrcpy (err, e.err);
	line = e.line;
	col = e.col;
	searchCol = e.searchCol;
	fileMatch = e.fileMatch;
}

TEICollection *eList = NULL;

static void initEList (void)
{
	if (eList == NULL) eList = new TEICollection;
}

static int findLngID (const char *fn)
{
	for (unsigned i = 0; i < langColl.getCount (); i++)
	{
		TLang *lng = (TLang *) (langColl[i]);
		if (matched (lng->mask, fn)) return (i);
	}

	return (-1);
}

static int eListInsert (EditorInfo *ei)
{
	int newFile = !isFile (ei->FileName);
	return (eList->insert (ei->EditorID, findLngID (ei->FileName), ei->FileName, newFile));
}

static FarMenuItemEx	*compilerOut = NULL;
static TCollection		*errColl = NULL;
static int						compilerOutN = 0;
static int						compilerOutP = -1;

struct TErrData
{
	char	fn[NM];
	int		line, col, msgCount;
	char	message[64][64];
	char	colSearch[MAX_REG_LEN];
};

static bool showErrorMsgBox (int res)
{
	int n = (int) compilerOut[res].UserData - 1;
	if (errColl && (n >= 0) && (n < (int) errColl->getCount ()))
	{
		TErrData		*err = (TErrData *) ((*errColl)[n]);
		const char	*MsgItems[] =
		{
			cmd, err->fn, "\x01", err->message[0], err->message[1], err->message[2], err->message[3], err->message[4],
			err->message[5], err->message[6], err->message[7], err->message[8], err->message[9], err->message[10],
			err->message[11], err->message[12], err->message[13], err->message[14], err->message[15], err->message[16],
			err->message[17], err->message[18], err->message[19]
		};
		redraw ();
		Info.Message (Info.ModuleNumber, FMSG_MB_OK | FMSG_LEFTALIGN, NULL, MsgItems, err->msgCount + 3, 0);
		return (true);
	}

	return (false);
}

static void jumpToError (EditorInfo *ei, int aj, bool showMsgBox)
{
	if (compilerOut && (aj >= 0) && (aj <= compilerOutN))
	{
		int n = (int) compilerOut[aj].UserData - 1;
		if (errColl && (n >= 0) && (n < (int) errColl->getCount ()))
		{
			TErrData	*err = (TErrData *) ((*errColl)[n]);
			int				nLine = -1, nCol = err->col, sLine = -1;
			if (err->line > 0) nLine = err->line - 1;
			if (err->col == 0) nCol--;

			char	errfn[NM];
			fExpand (lstrcpy (errfn, err->fn ? err->fn : ""));

			WindowInfo	wi;
			bool				found = false;
			int					n = Info.AdvControl (Info.ModuleNumber, ACTL_GETWINDOWCOUNT, NULL);
			for (int i = -1; i < n; i++)
			{
				wi.Pos = i;
				Info.AdvControl (Info.ModuleNumber, ACTL_GETWINDOWINFO, (void *) &wi);
				if (wi.Type == WTYPE_EDITOR)
				{
					if (!FSF.LStricmp (errfn, wi.Name))
					{
						if (i != -1)
						{
							Info.AdvControl (Info.ModuleNumber, ACTL_SETCURRENTWINDOW, (void *) i);
							Info.AdvControl (Info.ModuleNumber, ACTL_COMMIT, NULL);
						}

						sLine = nLine;
						found = true;
						break;
					}
				}
			}

			if (!found)
				Info.Editor (err->fn, NULL, 0, 0, -1, -1, EF_NONMODAL | EF_ENABLE_F6 | EF_IMMEDIATERETURN, err->line, nCol);
			EditorSetPosEx (ei, sLine, nCol, err->colSearch);
			if (showMsgBox) showErrorMsgBox (aj);
		}
	}
}

static void showCompileOut (EditorInfo *ei)
{
	if (compilerOut)
	{
		FarMenuItemEx *errOut = NULL;
		bool					ErrOnly = (filterring != 0);
		int						ErrCount = 0;
		int						Code, BreakKeys[] = { VK_RETURN, VK_F4, VK_F3, VK_MULTIPLY, 0 };
		if (compilerOutP >= 0 && compilerOutP < compilerOutN)
		{
			for (int i = 0; i < compilerOutN; i++) compilerOut[i].Flags &= ~MIF_SELECTED;
			compilerOut[compilerOutP].Flags |= MIF_SELECTED;
		}

		for (int i = 0; i < compilerOutN; i++)
		{
			if (!(compilerOut[i].Flags & MIF_DISABLE)) ErrCount++;
		}

		if (ErrCount > 0)
		{
			int j = 0;
			errOut = new FarMenuItemEx[ErrCount];
			for (int i = 0; i < compilerOutN; i++)
			{
				if (!(compilerOut[i].Flags & MIF_DISABLE)) errOut[j++] = compilerOut[i];
			}
		}

		for (;;)
		{
			compilerOutP = Info.Menu
				(
					Info.ModuleNumber,
					-1,
					-1,
					0,
					FMENU_WRAPMODE | FMENU_SHOWAMPERSAND | FMENU_USEEXT,
					cmd,
					GetMsg (MMenuBottom),
					NULL,
					BreakKeys,
					&Code,
					(ErrOnly) ? (const FarMenuItem *) errOut : (const FarMenuItem *) compilerOut,
					(ErrOnly) ? ErrCount : compilerOutN
				);
			if (ErrOnly)
			{
				int j = 0;
				for (int i = 0; i < compilerOutN; i++)
				{
					if (!(compilerOut[i].Flags & MIF_DISABLE))
						if (j == compilerOutP)
						{
							compilerOutP = i;
							break;
						}
						else
							j++;
				}
			}

			if (compilerOutP < 0) break;
			if (Code == 2)
			{
				if (compilerOutP <= compilerOutN && showErrorMsgBox (compilerOutP))
				{
					for (int i = 0; i < compilerOutN; i++) compilerOut[i].Flags &= ~MIF_SELECTED;
					compilerOut[compilerOutP].Flags |= MIF_SELECTED;
				}
			}
			else if (Code == 3)
			{
				ErrOnly = !ErrOnly;
			}
			else
			{
				jumpToError (ei, compilerOutP, false);
				break;
			}
		}

		if (ErrCount > 0) delete[] errOut;
	}
}

static bool SaveAll ()
{
	bool				res = true;
	WindowInfo	wi;
	wi.Pos = -1;
	Info.AdvControl (Info.ModuleNumber, ACTL_GETWINDOWINFO, (void *) &wi);

	int home = wi.Pos;
	int n = Info.AdvControl (Info.ModuleNumber, ACTL_GETWINDOWCOUNT, NULL);
	for (int i = 0; i < n; i++)
	{
		wi.Pos = i;
		Info.AdvControl (Info.ModuleNumber, ACTL_GETWINDOWINFO, (void *) &wi);
		if (wi.Type == WTYPE_EDITOR && wi.Modified)
		{
			if (Info.AdvControl (Info.ModuleNumber, ACTL_SETCURRENTWINDOW, (void *) i))
			{
				Info.AdvControl (Info.ModuleNumber, ACTL_COMMIT, NULL);
				if (!Info.EditorControl (ECTL_SAVEFILE, NULL)) res = false;
			}
			else
				res = false;
		}
	}

	Info.AdvControl (Info.ModuleNumber, ACTL_SETCURRENTWINDOW, (void *) home);
	Info.AdvControl (Info.ModuleNumber, ACTL_COMMIT, NULL);
	return (res);
}

static bool findBaseFile (const char *file, char *baseFile)
{
	bool	found = false;
	char	testFile[NM];
	int		n = Info.AdvControl (Info.ModuleNumber, ACTL_GETWINDOWCOUNT, NULL);
	GetCurrentDirectory (sizeof (testFile), testFile);
	FSF.AddEndSlash (testFile);
	lstrcat (testFile, FSF.PointToName (file));
	if (isFile (testFile))
	{
		if (baseFile) lstrcpy (baseFile, testFile);
		found = true;
	}

	if (!found)
	{
		for (int i = -1; i < n; i++)
		{
			WindowInfo	wi;
			wi.Pos = i;
			Info.AdvControl (Info.ModuleNumber, ACTL_GETWINDOWINFO, (void *) &wi);
			if (wi.Type == WTYPE_EDITOR)
			{
				char	*p = (char *) lstrrchr (lstrcpy (testFile, wi.Name), '\\');
				if (p)
				{
					lstrcpy (p + 1, FSF.PointToName (file));
					if (isFile (testFile))
					{
						if (baseFile) lstrcpy (baseFile, testFile);
						found = true;
						break;
					}
				}
			}
		}
	}

	return (found);
}

static bool validMenuItem (const char *fn, TExec *e)
{
	char	baseFile[NM] = "";
	bool	enable = true;
	if (e->enable[0])
	{
		makeCmdLine (false, baseFile, e->enable, fn);
		enable = isFile (baseFile);
		if (!enable && e->searchBase) enable = findBaseFile (baseFile, NULL);
	}

	if (enable && e->disable[0]) enable = !isFile (makeCmdLine (false, baseFile, e->disable, fn));
	return (enable);
}

static char *makeTitle (const char *line, size_t size, const char *fn)
{
	static char temp[MAX_STR_LEN];
	makeCmdLine (false, temp, line, fn);
	FSF.TruncStr (temp, size);
	return (temp);
}

static bool parseError (TLang *lng, const char *compiler, const char *fn, char *line, unsigned ci, int &aj)
{
	int				lineBounds[2], colBounds[2], fileBounds[2];
	TCompiler *e = NULL;
	bool			found = false, res = false;
	for (unsigned i = 0; i < lng->compilerColl.getCount (); i++)
	{
		e = (TCompiler *) (lng->compilerColl[i]);
		if (e && e->err[0] && !FSF.LStricmp (compiler, e->title))
		{
			int		bounds[3][2] = { { 0, 0 }, { 0, 0 }, { 0, 0 } };
			int		pos[3] = { e->line, e->col, e->fileMatch };
			char	lineRE[MAX_REG_LEN];
			lstrcpyn (lineRE, line, MAX_REG_LEN);
			lineRE[MAX_REG_LEN - 1] = 0;
			if (strMatch (lineRE, e->err, "/", ".*/i", 3, bounds, pos))
			{
				for (int j = 0; j < 2; j++)
				{
					lineBounds[j] = bounds[0][j];
					colBounds[j] = bounds[1][j];
					fileBounds[j] = bounds[2][j];
				}

				found = true;
				break;
			}
		}
	}

	if (found && e && e->err[0])
	{
		int len, lineNo = -1, colNo = -1;
		compilerOut[ci].Flags = 0;

		char	fileName[NM], colSearch[MAX_REG_LEN] = "";
		if (e->line >= 0)
		{
			len = lineBounds[1] - lineBounds[0];
			lstrcpyn (fileName, line + lineBounds[0], len + 1);
			fileName[len] = 0;
			if (len) lineNo = FSF.atoi (fileName);
		}

		if (e->col >= 0)
		{
			len = colBounds[1] - colBounds[0];
			lstrcpyn (fileName, line + colBounds[0], len + 1);
			fileName[len] = 0;
			if (len)
			{
				if (e->searchCol)
				{
					colNo = -2;
					lstrcpy (colSearch, fileName);
				}
				else
					colNo = FSF.atoi (fileName);
			}
		}

		if (e->fileMatch >= 0)
		{
			len = fileBounds[1] - fileBounds[0];
			lstrcpyn (fileName, line + fileBounds[0], len + 1);
			fileName[len] = 0;
			if (!len) lstrcpy (fileName, fn);
		}
		else
			lstrcpy (fileName, fn);

		TErrData	*errData = new TErrData;
		fExpand (lstrcpy (errData->fn, fileName));

		char	*p = line;
		errData->msgCount = 0;
		for (;;)
		{
			char	brk[260];
			Info.AdvControl (Info.ModuleNumber, ACTL_GETSYSWORDDIV, brk);

			char	*pb = errData->message[errData->msgCount];
			lstrcpyn (pb, p, 64);
			pb[63] = 0;
			if (lstrpbrk (pb, brk))
			{
				char	*pp = lstrchr (pb, 0);
				if (pp > pb)
				{
					while (!lstrchr (brk, *(--pp)))
						if (pp <= pb) break;
					pp[1] = 0;
				}
			}

			p += lstrlen (pb);
			FSF.Trim (pb);
			if (!*pb)
			{
				lstrcpy (pb, "\x01");
				errData->msgCount++;
				break;
			}

			errData->msgCount++;
		}

		errData->line = lineNo;
		errData->col = colNo;
		lstrcpy (errData->colSearch, colSearch);
		errColl->insert (errData);
		compilerOut[ci].UserData = errColl->getCount ();
		if (aj < 0)
		{
			compilerOut[ci].Flags = MIF_SELECTED;
			aj = (int) ci;
		}

		res = true;
	}

	return (res);
}

static bool runCompiler (EditorInfo *ei, TLang *lng, const char *fn, const char *title, TExec *e)
{
	char	cwd[NM] = "", baseFile[NM] = "";
	if (e->enable[0])
	{
		char	temp[NM];
		makeCmdLine (false, baseFile, e->enable, fn);

		bool	enable = isFile (baseFile);
		if (!enable && e->searchBase && findBaseFile (baseFile, temp)) lstrcpy (baseFile, temp);
	}

	if (!makeCmdLine (true, cmd, e->cmd, fn, makeTitle (title, 30, fn))) return (false);
	FSF.ExpandEnvironmentStr (cmd, cmd, sizeof (cmd));
	if (e->saveType != saNone)
	{
		if (e->saveType == saCurrent)
		{
			WindowInfo	wi;
			wi.Pos = -1;
			Info.AdvControl (Info.ModuleNumber, ACTL_GETWINDOWINFO, (void *) &wi);
			if (wi.Type == WTYPE_EDITOR && wi.Modified)
				if (!Info.EditorControl (ECTL_SAVEFILE, NULL)) return (false);
		}
		else
		{
			if (!SaveAll ())
			{
				const char	*MsgItems[] = { GetMsg (MTitle), GetMsg (MSaveError), GetMsg (MContinue) };
				if (Info.Message (Info.ModuleNumber, FMSG_MB_YESNO, NULL, MsgItems, dimOf (MsgItems), 0)) return (false);
			}
		}
	}

	if (e->cd != cdNone)
	{
		char	newDir[NM] = "", *p;
		if (e->cd == cdFile)
			p = (char *) lstrrchr (lstrcpy (newDir, fn), '\\');
		else if ((e->cd == cdBase) && *baseFile)
			p = (char *) lstrrchr (lstrcpy (newDir, baseFile), '\\');
		if (p) *p = 0;
		if (*newDir)
		{
			GetCurrentDirectory (sizeof (cwd), cwd);
			SetCurrentDirectory (newDir);
		}
	}

	if (errColl)
		errColl->removeAll ();
	else
		errColl = new TCollection;

	TCollection compilerColl;
	compilerOutP = -1;
	execute (&compilerColl, cmd, e->echo);

	unsigned	n = compilerColl.getCount ();
	if (n)
	{
		int nErr = 0, aj = -1;
		compilerOut = new FarMenuItemEx[compilerOutN = n + 1];
		for (unsigned i = 0; i < n; i++)
		{
			char	*line = (char *) (compilerColl[i]);
			lstrcpyn (compilerOut[i].Text.Text, line, sizeof (compilerOut[i].Text.Text));
			compilerOut[i].Text.Text[sizeof (compilerOut[i].Text.Text) - 1] = 0;
			compilerOut[i].Flags = MIF_DISABLE;
			compilerOut[i].UserData = 0;

			char	compiler[MAX_STR_LEN];
			lstrcpy (compiler, e->compiler);

			char	*pp, *p = compiler;
			bool	found = false;
			while ((pp = lstrchr (p, ',')) != NULL)
			{
				*pp = 0;
				if (parseError (lng, p, fn, line, i, aj))
				{
					nErr++;
					found = true;
				}

				p = pp + 1;
				if (found) break;
			}

			if (!found && parseError (lng, p, fn, line, i, aj)) nErr++;
		}

		if ((aj >= 0) && ((e->jumpType == jtFirst)) || ((e->jumpType == jtSmart) && (nErr == 1)))
			jumpToError (ei, aj, true);
		else
		{
			if (aj < 0)
			{
				if (compilerOut[n - 1].Text.Text[0] == 0)
				{
					compilerOutN--;
					n--;
				}

				compilerOut[n].Text.Text[0] = 0;
				compilerOut[n].Flags = MIF_SELECTED;
			}
			else
				compilerOutN--;
			if ((e->jumpType == jtMenu) || (e->jumpType == jtSmart)) showCompileOut (ei);
		}
	}
	else if (compilerOut)
	{
		delete[] compilerOut;
		compilerOut = NULL;
	}

	compilerColl.done ();
	if (*cwd) SetCurrentDirectory (cwd);
	return (true);
}

static bool sep = true;

static void addToCompilerMenu (const char *line, FarMenuItemEx *amenu, int &i, DWORD j, const char *fn, DWORD AccelKey)
{
	FSF.sprintf (amenu[i].Text.Text, "%s", makeTitle (line, sizeof (amenu[i].Text.Text) - 1, fn));
	amenu[i].UserData = j;
	amenu[i].Flags = 0;
	amenu[i].AccelKey = AccelKey;
	if (amenu[i].Text.Text[0])
	{
		i++;
		sep = false;
	}
	else
	{
		if (!sep)
		{
			amenu[i].Flags = MIF_SEPARATOR;
			i++;
		}

		sep = true;
	}
}

static void CompilerMenu (EditorInfo *ei, const char *Name, int lang)
{
	char	top[NM], FileName[NM];
	lstrcpy (FileName, Name);
	lstrcpy (top, GetMsg (MTitle));

	TLang			*lng = (TLang *) (langColl[lang]);
	unsigned	ec = 0;
	if (lng) ec = lng->execColl.getCount ();
	if ((!ec) && (!outputmenu)) return ;
	if (autocompile)
	{
		int i, count = 0;
		if (ec)
		{
			for (unsigned j = 0; j < ec; j++)
			{
				TExec *ex = ((TExec *) lng->execColl[j]);
				if (validMenuItem (FileName, ex))
				{
					i = j;
					count++;
				}
			}
		}

		if (count == 1)
			if (lng)
			{
				TExec *e = (TExec *) lng->execColl[i];
				runCompiler (ei, lng, FileName, e->title, e);
				return ;
			}
	}

	FarMenuItemEx *amenu = new FarMenuItemEx[ec + 2];
	if (amenu)
	{
		const DWORD scId = -1;
		int					i = 0;
		sep = true;
		if (ec)
		{
			for (unsigned j = 0; j < ec; j++)
			{
				TExec *ex = ((TExec *) lng->execColl[j]);
				if (validMenuItem (FileName, ex)) addToCompilerMenu (ex->title, amenu, i, j, FileName, 0);
			}

			addToCompilerMenu ("", amenu, i, 0, FileName, 0);
		}

		addToCompilerMenu (GetMsg (MShowOutput), amenu, i, scId, FileName, KEY_MULTIPLY);
		if (!compilerOut) amenu[i - 1].Flags = MIF_DISABLE;
		for (int k = 0; k < i; k++)
			if (!(amenu[k].Flags & (MIF_DISABLE | MIF_SEPARATOR)))
			{
				amenu[k].Flags |= MIF_SELECTED;
				break;
			}

		int res = Info.Menu
			(
				Info.ModuleNumber,
				-1,
				-1,
				0,
				FMENU_WRAPMODE | FMENU_AUTOHIGHLIGHT | FMENU_USEEXT,
				top,
				NULL,
				NULL,
				NULL,
				NULL,
				(const FarMenuItem *) amenu,
				(outputmenu) ? i : i - 2
			);
		if (res != -1)
		{
			DWORD id = amenu[res].UserData;
			if (id == scId)
				showCompileOut (ei);
			else if (lng)
			{
				TExec *e = (TExec *) lng->execColl[id];
				runCompiler (ei, lng, FileName, e->title, e);
			}
		}

		delete[] amenu;
	}
}

static void SelectCompiler (TEInfo *te)
{
	EditorInfo	ei;
	Info.EditorControl (ECTL_GETINFO, &ei);
	CompilerMenu (&ei, ei.FileName, te->lang);
}

static void ShowOutput ()
{
	EditorInfo	ei;
	Info.EditorControl (ECTL_GETINFO, &ei);
	showCompileOut (&ei);
}
