static int inComment (TLang *lng, const char *text_line, unsigned pos)
{
	if (text_line)
	{
		for (unsigned i = 0; i < lng->commentColl.getCount (); i++)
		{
			TComment	*tmpc = (TComment *) (lng->commentColl[i]);
			char			*line = (char *) malloc (lstrlen (text_line) + 1);

			if (line)
			{
				lstrcpy (line, text_line);
				if (lng->ignoreCase) StrLower (line); /*!?*/

				int		bounds[2], zero = 0;
				bool	blMatched = strMatch (line, tmpc->mask, "/", lng->ignoreCase ? "/i" : "/", 1, &bounds, &zero);
				free (line);

				if (blMatched && pos >= (unsigned) bounds[0] && pos <= (unsigned) bounds[1])
				{
					return (1);
				}
			}
		}
	}

	return (0);
}

static int FindIndent (char c, TLang *lng, char *str, int foundBounds[2])
{
	for (unsigned j = 0; j < lng->indentColl.getCount (); j++)
	{
		TIndent *tmpi = (TIndent *) (lng->indentColl[j]);
		char		ic = ((TIndent *) (lng->indentColl[j]))->immChar;
		if (lng->ignoreCase) ic = (char) FSF.LLower (ic);
		if ((c == '\0') || (c == ic))
		{
			char	*line = (char *) malloc (lstrlen (tmpi->mask) + 1);

			if (line)
			{
				lstrcpy (line, tmpi->mask);

				int		bounds[2];
				bool	blMatched = strMatch (str, line, "/^", lng->ignoreCase ? "$/i" : "$/", 1, &bounds, &tmpi->start);
				free (line);

				if (blMatched)
				{
					foundBounds[0] = bounds[0];
					foundBounds[1] = bounds[1];
					return (j);
				}
			}
		}
	}

	return (-1);
}

static int FindBrackets (TLang *lng, char *str, char **foundBrackets)
{
	int foundBracketsNum = 0;

	for (unsigned j = 0; j < lng->indentColl.getCount (); j++)
	{
		TIndent *tmpi = (TIndent *) (lng->indentColl[j]);
		char		*line = (char *) malloc (lstrlen (tmpi->mask) + 1);

		if (line)
		{
			lstrcpy (line, tmpi->mask);
			if (lng->ignoreCase) StrLower (line) /* ?! */ ;

			int		bounds[2];
			bool	blMatched = strMatch (str, line, "/^", lng->ignoreCase ? "$/i" : "$/", 1, &bounds, &tmpi->start);
			free (line);

			if (blMatched)
			{
				if (tmpi->indent[0] != 0xFFFF && tmpi->BracketsMode)
				{
					TBracket	*br = (TBracket *) (lng->bracketColl[tmpi->bracketLink]);
					foundBrackets[foundBracketsNum++] = br->open;
				}
			}
		}
	}

	return (foundBracketsNum);
}

static char *LookForOpenBracket (EditorInfo *ei, TEditorPos &p, TLang *lng, char **openBr, int nOpenBr)
{
	static char							*c, buff[MAX_STR_LEN];
	static EditorGetString	gs;
	int		ret = -1;
	char	*closeReg = lng->ignoreCase ? "$/i" : "$/";
	while (--p.Row >= 0)
	{
		EditorSetPos (p);
		EditorGetStr (&gs);

		if (gs.StringLength)
		{
			if ((c = FirstNonSpace (gs.StringText)) != NULL)
			{
				Editor2OEM (ei, FSF.Trim (lstrcpy (buff, c)));
				if (lng->ignoreCase) StrLower (buff);
				for (int i = 0; i < nOpenBr; i++)
				{
					if (strMatch (buff, openBr[i], "/^", closeReg, NULL, 0))
					{
						ret = c - gs.StringText;
						break;
					}
				}

				if (ret != -1) break;

				unsigned	n = lng->bracketColl.getCount ();
				{
					for (unsigned i = 0; i < n; i++)
					{
						TBracket	*br = (TBracket *) (lng->bracketColl[i]);
						if (strMatch (buff, br->close, "/^", closeReg, NULL, 0))
						{
							char	*result = NULL, **br4look = new char *[n];
							int		nbr4look = FindBrackets (lng, buff, br4look);
							if (nbr4look) result = LookForOpenBracket (ei, p, lng, br4look, nbr4look);
							delete[] br4look;
							if (result) break;
							return (NULL);
						}
					}
				}
			}
		}
	}

	if (ret != -1)
	{
		memcpy (buff, gs.StringText, ret);
		buff[ret] = 0;
		return (buff);
	}

	return (NULL);
}

static char *FindOpenBracket (EditorInfo *ei, TLang *lng, char **openBr, int n)
{
	TEditorPos	p, o = EditorGetPos ();
	p.Default ();
	p.Row = o.Row;

	char	*ret = LookForOpenBracket (ei, p, lng, openBr, n);
	EditorSetPos (o);
	return (ret);
}

static char *prevStringIndent (void)
{
	static char			buff[MAX_STR_LEN];
	EditorGetString gs;
	int							ret = 0;
	TEditorPos			p, o = EditorGetPos ();
	p.Default ();
	p.Row = o.Row;
	while (--p.Row >= 0)
	{
		EditorSetPos (p);
		EditorGetStr (&gs);
		if (gs.StringLength)
		{
			char	*c = FirstNonSpace (gs.StringText);
			if (c)
			{
				ret = c - gs.StringText;
				break;
			}
		}
	}

	EditorSetPos (o);
	memcpy (buff, gs.StringText, ret);
	buff[ret] = 0;
	return (buff);
}

static int shiftLine (int i)
{
	TEditorPos	pos = EditorGetPos ();
	pluginBusy = 1;
	if (i > 0)
		for (int j = 0; j < i; j++)
			EditorProcessFARKey (KEY_TAB);
	else if (i < 0)
	{
		EditorInfo	ei;
		Info.EditorControl (ECTL_GETINFO, &ei);

		int ind = -i * ei.TabSize;
		for (int j = 0; j < ind; j++) EditorProcessFARKey (KEY_BS);
	}

	TEditorPos	pos2 = EditorGetPos ();
	pluginBusy = 0;
	return (pos2.Col - pos.Col);
}

static int TryIndent
(
	EditorInfo	*ei,
	int					ret,
	TLang				*lng,
	int					j,
	char				*str,
	char				*realStr,
	int					oldIndent,
	int					foundBounds[2]
)
{
	TIndent *tmpi = (TIndent *) (lng->indentColl[j]);
	int			iCurr = tmpi->indent[0];
	int			iNext = tmpi->indent[1];
	if (iCurr != 0xFFFF)
	{
		bool				processIndent = true;
		char				*opencol = NULL;
		TEditorPos	pos = EditorGetPos ();
		if (*(tmpi->relative))
		{
			char	*rel = tmpi->relative;
			if ((opencol = FindOpenBracket (ei, lng, &rel, 1)) == NULL) processIndent = false;
		}
		else if (tmpi->BracketsMode)
		{
			unsigned	n = lng->bracketColl.getCount ();
			char			**br4look = new char *[n];
			int				nbr4look = FindBrackets (lng, str, br4look);
			processIndent = false;
			if (nbr4look) opencol = FindOpenBracket (ei, lng, br4look, nbr4look);
			if (opencol) processIndent = true;
			delete[] br4look;
		}

		if (processIndent)
		{
			if (opencol == NULL) opencol = prevStringIndent ();

			size_t	olen = lstrlen (opencol);
			char		*ptr = new char[olen + lstrlen (str) + 1];
			lstrcat (lstrcpy (ptr, opencol), realStr);
			EditorSetStr (ptr);
			EditorSetPos (-1, olen);

			int sh = shiftLine (iCurr);
			EditorSetPos (-1, pos.Col + olen - oldIndent + sh);
		}
	}

	if (ret)
	{
		EditorProcessReturnKey (1, 0);
		if (tmpi->start)
		{
			pluginBusy = 1;
			for (int k = foundBounds[0]; k < foundBounds[1]; k++) EditorProcessKey (realStr[k]);
			pluginBusy = 0;
		}

		if (iNext != 0xFFFF) shiftLine (iNext);
	}
	else
		redraw ();
	return (IGNORE_EVENT);
}

static bool checkMultipleChoice (TLang *lng, TMacro * &mc, char *before, char *after, int len, int bounds[][2])
{
	const char	*menuPrefix = "\\~";
	unsigned		pl = lstrlen (menuPrefix);
	if ((unsigned) lstrlen (mc->MacroText) > pl && !FSF.LStrnicmp (mc->MacroText, menuPrefix, pl))
	{
		unsigned	lc = 0;
		char			*p = mc->MacroText;
		while (*p)
		{
			if (p[0] == '\\')
			{
				if (p[1] == '~')
				{
					lc++;
					p += 2;
				}
				else
				{
					p++;
					if (p[0] == '\0')
						break;
					else
						p++;
				}
			}
			else
				p++;
		}

		if (lc)
		{
			FarMenuItem *amenu = new FarMenuItem[lc];
			TMacro			**amacro = new TMacro *[lc];
			if (amenu)
			{
				unsigned	i;
				p = mc->MacroText;
				for (i = 0; ((i < lc) && (*p));)
				{
					TMacro	*tm;
					char		tmpTitle[MAX_STR_LEN];
					char		tmpBefore[MAX_STR_LEN];
					lstrcpy (tmpBefore, before);
					tmpBefore[lstrlen (tmpBefore) - len] = 0;

					char	*p1 = tmpBefore + lstrlen (tmpBefore);
					if ((p[0] == '\\') && (p[1] == '~')) p += 2;
					while (*p)
					{
						if (p[0] == '\\')
						{
							if (p[1] == '~')
							{
								p += 2;
								break;
							}
							else
							{
								p++;
								if (p[0] == '\0')
									break;
								else
								{
									*p1 = *p;
									p++;
									p1++;
								}
							}
						}
						else
						{
							if (*p == '=')
							{
								p++;
								*p1 = '\0';
								p1 = tmpTitle;
							}
							else
							{
								*p1 = *p;
								p++;
								p1++;
							}
						}
					}

					*p1 = '\0';
					tm = FindMacro (lng, tmpBefore, after, 0, NULL, bounds, true);
					lstrcpy (amenu[i].Text, tmpTitle);
					if (tm)
					{
						amacro[i] = tm;
						amenu[i].Selected = amenu[i].Checked = amenu[i].Separator = false;
						i++;
					}
				}

				int		res = Info.Menu
					(
						Info.ModuleNumber,
						-1,
						-1,
						0,
						FMENU_WRAPMODE,
						GetMsg (MSelectMacro),
						NULL,
						NULL,
						NULL,
						NULL,
						amenu,
						i
					);
				bool	retval = false;
				if (res != -1 && amacro[res])
				{
					mc = amacro[res];
					retval = true;
				}

				delete[] amenu;
				delete[] amacro;
				return (retval);
			}
		}
	}

	return (true);
}

static int TryMacro (EditorInfo *ei, TLang *lng, EditorGetString &gs, TEditorPos &ep, char *line, char expChar)
{
	if (expChar || !isCharSpace (gs.StringText[ep.Col - 1]))
	{
		if (gs.StringLength != 0)
		{
			int l = gs.StringLength;
			lstrcpyn (line, gs.StringText, l + 1);
			for (int i = l; i < MAX_STR_LEN; i++) line[i] = ' ';
			line[MAX_STR_LEN - 1] = '\0';
			Editor2OEM (ei, line);
		}

		if (!inComment (lng, line, ep.Col))
		{
			char	before[MAX_STR_LEN];
			char	*after = line + ep.Col;
			lstrcpy (before, line);

			int len = before[ep.Col] = 0;
			if (expChar)
			{
				before[ep.Col] = expChar;
				before[ep.Col + 1] = 0;
			}

			int			bounds[16][2];
			TMacro	*mc = FindMacro (lng, before, after, expChar, &len, bounds);
			if (mc && checkMultipleChoice (lng, mc, before, after, len, bounds))
			{
				if (expChar) len--;
				pluginBusy = 1;
				for (int i = 0; i < len; i++) EditorProcessFARKey (KEY_BS);
				pluginBusy = 0;
				RunMacro (mc, before, bounds);
				redraw ();
				return (1);
			}
		}
	}

	return (0);
}

int WINAPI _export ProcessEditorInput (const INPUT_RECORD *Rec)
{
	if
	(
		pluginStop ||
		pluginBusy ||
		(Rec->EventType != KEY_EVENT && Rec->EventType != 0x8001) ||
		Rec->Event.KeyEvent.bKeyDown == 0
	) return (PROCESS_EVENT);

	int vState = Rec->Event.KeyEvent.dwControlKeyState;
	if (scrollStop && (vState & SCROLLLOCK_ON)) return (PROCESS_EVENT);

	EditorInfo	ei;
	Info.EditorControl (ECTL_GETINFO, &ei);
	initEList ();

	int edId = eList->findID (ei.EditorID);
	if (edId == -1) edId = eListInsert (&ei);

	TEInfo	*te = (TEInfo *) (*eList)[edId];
	int			id = te->lang;
	if (id != -1)
	{
		TLang *lng = (TLang *) (langColl[id]);
		if (lng)
		{
			//! NEW FILE - START
			if (te->newFile)
			{
				te->newFile = 0;
				if (!Rec->Event.KeyEvent.wVirtualKeyCode)
				{
					if (lng->setCP)
					{
						EditorSetParameter	esp;
						esp.Type = ESPT_CHARTABLE;
						esp.Param.iParam = lng->setCP;
						Info.EditorControl (ECTL_SETPARAM, &esp);
					}

					//Execute init macros
					unsigned	l = 0;
					unsigned	n = 0;
					for (unsigned i = 0; i < lng->macroColl.getCount (); i++)
						if ((((TMacro *) (lng->macroColl[i]))->atStartup))
						{
							l = i;
							n++;
						}

					if (n == 1)
					{
						TMacro	*mm = (TMacro *) (lng->macroColl[l]);
						if (checkMultipleChoice (lng, mm, "", "", 0, NULL))
						{
							RunMacro (mm, NULL, NULL);
							redraw ();
						}
					}

					if (n > 1)
					{
						l = 0;

						FarMenuItem *pMenu = new FarMenuItem[n];
						for (unsigned i = 0; i < lng->macroColl.getCount (); i++)
						{
							if ((((TMacro *) (lng->macroColl[i]))->atStartup))
							{
								if (lstrcmp (((TMacro *) (lng->macroColl[i]))->Name, "") == 0)
									lstrcpy (pMenu[l].Text, GetMsg (MUnnamed));
								else
									lstrcpy (pMenu[l].Text, ((TMacro *) (lng->macroColl[i]))->Name);
								l++;
							}
						}

						unsigned	res = Info.Menu
							(
								Info.ModuleNumber,
								-1,
								-1,
								0,
								FMENU_WRAPMODE,
								GetMsg (MSelectMacro),
								NULL,
								NULL,
								NULL,
								NULL,
								pMenu,
								n
							);
						if (res != -1)
						{
							n = 0;
							for (unsigned i = 0; i < lng->macroColl.getCount (); i++)
								if ((((TMacro *) (lng->macroColl[i]))->atStartup))
								{
									if (n == res)
									{
										l = i;
										break;
									}

									n++;
								}

							TMacro	*mm = (TMacro *) (lng->macroColl[l]);
							if (checkMultipleChoice (lng, mm, "", "", 0, NULL))
							{
								RunMacro (mm, NULL, NULL);
								redraw ();
							}
						}

						delete[] pMenu;
					}

					return (IGNORE_EVENT);
				}
			} //! NEW FILE - END

			char	line[MAX_STR_LEN];
			for (int i = 0; i < MAX_STR_LEN; i++) line[i] = ' ';
			line[MAX_STR_LEN - 1] = '\0';

			TEditorPos	epos = EditorGetPos ();
			TMacro			*fm;

			//! MACRO EXISTS - START
			if ((fm = FindMacroKey (lng, Rec)) != NULL)
			{
				EditorGetString gs;
				EditorGetStr (&gs);

				if (gs.StringLength != 0)
				{
					int l = gs.StringLength;
					lstrcpyn (line, gs.StringText, l + 1);
					for (int i = l; i < MAX_STR_LEN; i++) line[i] = ' ';
					line[MAX_STR_LEN - 1] = '\0';
					Editor2OEM (&ei, line);
				}

				if (!inComment (lng, line, epos.Col))
				{
					char	before[MAX_STR_LEN], *after = line + epos.Col;
					lstrcpy (before, line);
					before[epos.Col] = 0;
					if (CheckMacroPos (lng, fm, before, after))
					{
						if (checkMultipleChoice (lng, fm, before, after, 0, NULL))
						{
							RunMacro (fm, NULL, NULL);
							redraw ();
							return (IGNORE_EVENT);
						}
					}
				}

				return (PROCESS_EVENT);
			} //! MACRO EXISTS - END

			int fKey = FSF.FarInputRecordToKey (Rec);
			vState &= (SHIFT_PRESSED | RIGHT_ALT_PRESSED | LEFT_ALT_PRESSED | RIGHT_CTRL_PRESSED | LEFT_CTRL_PRESSED);

			int spORret = (fKey == KEY_ENTER || fKey == defExpandFKey);
			if (!spORret && (vState &~SHIFT_PRESSED)) return (PROCESS_EVENT);

			int		inImm = 0, inImmExp = 0;
			char	vChar = Rec->Event.KeyEvent.uChar.AsciiChar;
			if (!spORret)
			{
				if (lng->ignoreCase)
				{
					char	lwrStr[MAX_REG_LEN], c = (char) (FSF.LLower (vChar));
					inImm = lstrchr (StrLower (lstrcpy (lwrStr, lng->imm)), c) != NULL;
					inImmExp = lstrchr (StrLower (lstrcpy (lwrStr, lng->immExp)), c) != NULL;
				}
				else
				{
					inImm = lstrchr (lng->imm, vChar) != NULL;
					inImmExp = lstrchr (lng->immExp, vChar) != NULL;
				}
			}

			if (!(spORret || (vChar && (inImm || inImmExp)))) return (PROCESS_EVENT);
			if ((fKey == defExpandFKey) || inImmExp)
			{
				EditorGetString gs;
				EditorGetStr (&gs);
				return (TryMacro (&ei, lng, gs, epos, line, inImmExp ? vChar : '\0') ? IGNORE_EVENT : PROCESS_EVENT);
			}
			else
			{
				int							ret = PROCESS_EVENT;
				EditorGetString gs;
				EditorGetStr (&gs);
				if (gs.StringLength != 0)
				{
					int l = gs.StringLength;
					lstrcpyn (line, gs.StringText, l + 1);
					for (int i = l; i < MAX_STR_LEN; i++) line[i] = ' ';
					line[MAX_STR_LEN - 1] = '\0';
					Editor2OEM (&ei, line);
				}

				if (!inComment (lng, line, epos.Col))
				{
					if (!spORret)
					{
						pluginBusy = 1;
						EditorProcessKey (vChar);
						pluginBusy = 0;
						ret = IGNORE_EVENT;
						EditorGetStr (&gs);
						if (gs.StringLength != 0)
						{
							int l = gs.StringLength;
							lstrcpyn (line, gs.StringText, l + 1);
							for (int i = l; i < MAX_STR_LEN; i++) line[i] = ' ';
							line[MAX_STR_LEN - 1] = '\0';
							Editor2OEM (&ei, line);
						}
					}

					char	*pstrText = FirstNonSpace (gs.StringText);
					if (pstrText == NULL) return (ret);

					int		lstr = lstrlen (pstrText);
					char	*str = (char *) malloc (lstr + 1);
					if (str)
					{
						int spacesCount = pstrText - gs.StringText;
						strncpy (str, pstrText, lstr + 1);
						FSF.Trim (line);

						if (lng->ignoreCase)
						{
							StrLower (line);
							vChar = (char) FSF.LLower (vChar);
						}

						int bounds[2];
						int j = FindIndent (spORret ? '\0' : vChar, lng, line, bounds);
						if (j != -1) ret = TryIndent (&ei, spORret, lng, j, line, str, spacesCount, bounds);
						free (str);
					}

					return (ret);
				}
			}
		}
	}

	return (PROCESS_EVENT);
}
