//====================================================================================================================
// Point to filename in a full path
//====================================================================================================================
static char *Point2FileName (char *szPath)
{
	int i = lstrlen (szPath);
	while (true)
	{
		i--;
		if (i < 0) break;
		if (szPath[i] == '\\') break;
	}

	i++;
	return (szPath + i);
}

//====================================================================================================================
// Delete fileext from full path
//====================================================================================================================
static char *Point2FileExt (char *szPath)
{
	int l;
	int i = l = lstrlen (szPath);
	while (true)
	{
		i--;
		if (i < 0) break;
		if (szPath[i] == '.') break;
		if (szPath[i] == '\\')
		{
			i = l;
			szPath[i] = '.';
			break;
		}
	}

	i++;
	return (szPath + i);
}

static char *addExt (char *path, char *defExt)
{
	int n = lstrlen (path) - 1;
	for (int i = n; i >= 0; i--)
	{
		if (path[i] == '.')
		{
			lstrcpy (path + i, defExt);
			break;
		}

		if ((path[i] == '\\') || (path[i] == ':') || (i == 0)) return (lstrcat (path, defExt));
	}

	return (path);
}

static bool IsFile (const char *szFileName)
{
	DWORD d = GetFileAttributes (szFileName);
	return (((HANDLE) d != INVALID_HANDLE_VALUE) && !(d & FILE_ATTRIBUTE_DIRECTORY));
}

static void squeeze (char *path)
{
	char	*dest = path;
	char	*src = path;
	while (*src != 0)
	{
		if (*src != '.')
			*dest++ = *src++;
		else
		{
			src++;
			if (*src == '.')
			{
				src += 2;
				dest--;
				while (*--dest != '\\');
				dest++;
			}
			else
			{
				src++;
				dest += 2;
			}
		}
	}

	*dest = 0;
}

enum enSplitType
{
	enDRIVE			= 1,
	enDIRECTORY = 2,
	enFILENAME	= 4,
	enEXTENTION = 8
};

static int fnSplit (const char *path, char *drive, char *dir, char *file, char *ext)
{
	int		flag = *drive = *dir = *file = *ext = 0;
	char	ch, *p = lstrchr ((char *) path, ':');
	if (p)
	{
		ch = p[1];
		p[1] = 0;
		lstrcpy (drive, path);
		flag |= enDRIVE;
		p[1] = ch;
		p++;
	}
	else
		p = (char *) path;

	char	*dp = lstrrchr (p, '\\');
	if (dp)
	{
		ch = dp[1];
		dp[1] = 0;
		lstrcpy (dir, p);
		dp[1] = ch;
		p = dp + 1;
		flag |= enDIRECTORY;
	}

	dp = lstrchr (p, '.');
	if (dp)
	{
		lstrcpy (ext, dp);
		flag |= enEXTENTION;
		ch = *dp;
		*dp = 0;
		lstrcpy (file, p);
		*dp = ch;
	}
	else
		lstrcpy (file, p);
	if (*file) flag |= enFILENAME;
	return (flag);
}

static inline char *fnMerge (char *path, char *drive, char *dir, char *file, char *ext)
{
	FSF.AddEndSlash (lstrcat (lstrcpy (path, drive), dir));
	return (lstrcat (lstrcat (path, file), ext));
}

static char *fExpand (char *rpath, char *defPath = NULL)
{
	static char path[NM], drive[NM], dir[NM], file[NM], ext[NM];
	static char defDrive[4], defDir[NM];
	if (defPath)
		lstrcpy (path, defPath);
	else
		GetCurrentDirectory (sizeof (path), path);
	FSF.AddEndSlash (path);
	fnSplit (path, defDrive, defDir, file, ext);

	int flags = fnSplit (rpath, drive, dir, file, ext);
	if ((flags & enDRIVE) == 0) lstrcpy (drive, defDrive);
	CharUpperBuff (drive, 1);
	if ((flags & enDIRECTORY) == 0 || ((*dir != '\\') && (*dir != '/')))
	{
		static char curdir[NM];
		lstrcpy (curdir, dir);
		lstrcat (lstrcpy (dir, defDir), curdir);
	}

	squeeze (dir);
	fnMerge (path, drive, dir, file, ext);
	return (lstrcpy (rpath, path));
}

static char *getFile (char *path, char *fn)
{
	char		file[NM], *fileBuff = NULL;
	HANDLE	s = CreateFile
		(
			fExpand (lstrcpy (file, fn), path),
			GENERIC_READ,
			FILE_SHARE_READ,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			NULL
		);
	if (s != INVALID_HANDLE_VALUE)
	{
		unsigned	fileSize = GetFileSize (s, NULL);
		if (fileSize)
		{
			if ((fileBuff = new char[fileSize + 1]) != NULL)
			{
				DWORD rb = 0;
				if (ReadFile (s, fileBuff, fileSize, &rb, NULL) && rb)
					fileBuff[rb] = 0;
				else
				{
					delete[] fileBuff;
					fileBuff = NULL;
				}
			}
		}
	}

	CloseHandle (s);
	return (fileBuff);
}

static inline char *delBackslash (char *s)
{
	if (*s && lstrcmp (s, "\\"))
	{
		char	*b = lstrchr (s, 0) - 1;
		if (*b == '\\') *b = 0;
	}

	return (s);
}

static char *QuoteText (char *Str)
{
	if (*Str == '-' || *Str == '^' || lstrpbrk (Str, " &+,"))
	{
		unsigned	l = lstrlen (Str);
		for (int i = l; i > 0; i--) Str[i] = Str[i - 1];
		*Str = Str[l + 1] = '\"';
		Str[l + 2] = 0;
	}

	return (Str);
}

static inline char *quote (int doQuote, char *Str)
{
	return (doQuote ? QuoteText (Str) : Str);
}

static bool isFile (const char *s)
{
	WIN32_FIND_DATA ff;
	HANDLE					h = FindFirstFile (s, &ff);
	bool						res = h != INVALID_HANDLE_VALUE;
	FindClose (h);
	return (res);
}

static inline int matched (const char *mask, const char *value)
{
	if (lstrchr ((char *) mask, '\\'))
		return (FSF.ProcessName (mask, (char *) value, PN_CMPNAMELIST));
	else
		return (FSF.ProcessName (mask, (char *) value, PN_CMPNAMELIST | PN_SKIPPATH));
}
